// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

static double glob_wanted=0.0;
static int glob_use_cap=0;
static int ntricks_fast;

void set_light_for_voc(struct device *in,double Voc)
{
//<clean>
double clamp=100;
double step=10;
double e0;
double e1;
double deriv;

solve_light(in,&(in->mylight),in->Psun,0.0);

newton_sim_tpv(in);
printf("Looking for Voc=%le\n",Voc);
//sim_i(in,0.0);
e0=fabs(in->Vapplied-Voc);

in->Psun+=step;
solve_light(in,&(in->mylight),in->Psun,0.0);
newton_sim_tpv(in);
//sim_i(in,0.0);
e1=fabs(in->Vapplied-Voc);

deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Psun+=step;

int count=0;
int max=1000;

do
{
e0=e1;
newton_sim_tpv(in);
//sim_i(in,0.0);
e1=fabs(in->Vapplied-Voc);

deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Psun+=step;
in->Psun=fabs(in->Psun);
solve_light(in,&(in->mylight),in->Psun,0.0);

printf("%d %le %le %le\n",count,e1,in->Psun,in->Vapplied);
if (count>max) break;
count++;
}while(e1>1e-5);
printf("%le %le %le #ans\n",in->Vapplied,in->Psun/1000.0,get_extracted_np(in));
printf("%le %le #2ans\n",get_extracted_np(in),get_avg_recom(in)/(pow(get_extracted_np(in),2.0)));
//</clean>
}

void set_ntricks_fast(int val)
{
ntricks_fast=val;
}

void ramp_externalv(struct device *in,double from,double to)
{
double V=from;
double dV=0.1;
if ((to-from)<0.0) dV*=-1.0;
printf("dV=%le\n",dV);
printf("Ramping: from=%e to=%e\n",from,to);

if (fabs(to-from)<=fabs(dV)) return;

do
{
V+=dV;
printf("ramp: %lf %le %d\n",V,to,in->newton_cur);
sim_externalv(in,V);

if (fabs(in->Vapplied-to)<fabs(dV))
{
	break;
}

}while(1);

if (V!=to)
{
	V=to;
	sim_externalv(in,V);
}

return;
}

void ramp(struct device *in,double from,double to,double steps)
{
in->Vapplied=from;
double dV=0.1;
if ((to-from)<0.0) dV*=-1.0;
printf("dV=%le\n",dV);
printf("Ramping: from=%e to=%e\n",from,to);

if (fabs(to-from)<=fabs(dV)) return;

do
{
in->Vapplied+=dV;
printf("ramp: %lf %le %d\n",in->Vapplied,to,in->newton_cur);
solve_all(in);

//sim_externalv(in,in->cevoltage);

if (fabs(in->Vapplied-to)<fabs(dV))
{
//save_state(in,to);
	break;
}

}while(1);

if (in->Vapplied!=to)
{
	in->Vapplied=to;
	solve_all(in);
}

return;
}


void save_state(struct device *in,double to)
{
//<clean>
printf("Dumping state\n");
int i;
int band;
FILE *state;
state=fopena(in->outputpath,"./state.dat","w");

fprintf(state,"%le ",to);


for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(state,"%le %le %le ",in->phi[i],in->x[i],in->xp[i]);

	for (band=0;band<in->srh_bands;band++)
	{
		fprintf(state,"%le %le ",in->xt[i][band],in->xpt[i][band]);
	}

}
fclose(state);
//</clean>
}

int load_state(struct device *in,double voltage)
{
//<clean>
printf("Load state\n");
int i;
int band;
double vtest;
FILE *state;
state=fopena(in->outputpath,"./state.dat","r");
if (!state)
{
printf("State not found\n");
return FALSE;
}

fscanf(state,"%le",&vtest);
printf("%le %le",voltage,vtest);
if (vtest!=voltage)
{
printf("State not found\n");
return FALSE;
}
printf("Loading state\n");

in->Vapplied=vtest;

for (i=0;i<in->ymeshpoints;i++)
{
	fscanf(state,"%le %le %le ",&(in->phi[i]),&(in->x[i]),&(in->xp[i]));

	for (band=0;band<in->srh_bands;band++)
	{
		fscanf(state,"%le %le ",&(in->xt[i][band]),&(in->xpt[i][band]));
	}

}
fclose(state);
return TRUE;
//</clean>
}


//    ----=----
//    |       |-----
//    |       |    |
//    |       |   ---
//    |       SC
//    |       |   ---
//    |       |    |
//  Vtot      |-----
//    |       R
//    |	      |
//    ----=----
void newton_aux_celiv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double C=in->C;
if (glob_use_cap==FALSE) C=0.0;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-(V+((in->Rcontact+in->Rshort_celiv)*(i0+V/in->Rshunt+C*(V-in->Vapplied_last)/in->dt)));
*didv=-(1.0+((in->Rcontact+in->Rshort_celiv)*(didv0+1.0/in->Rshunt+C*(1.0)/in->dt)));
*didphi=-((in->Rcontact+in->Rshort_celiv)*didphi0);
*didxil=-((in->Rcontact+in->Rshort_celiv)*didxil0);
*didxipl=-((in->Rcontact+in->Rshort_celiv)*didxipl0);
*didphir=-((in->Rcontact+in->Rshort_celiv)*didphir0);
*didxir=-((in->Rcontact+in->Rshort_celiv)*didxir0);
*didxipr=-((in->Rcontact+in->Rshort_celiv)*didxipr0);
//</clean>
return;
}


double newton_celiv(struct device *in,double Vtot,int usecap)
{
//<clean>
double C=in->C;
in->newton_cur=TRUE;
solver_realloc(in);
glob_wanted=Vtot;
glob_use_cap=usecap;
in->newton_aux=&newton_aux_celiv;
solve_all(in);
if (glob_use_cap==FALSE) C=0.0;
return get_I(in)+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt;
//</clean>
}

//    ----  ----
//    |        |
//    |        |
//   SC        R
//    |        |
//    |	       |
//    ----  ----
//So we hand the solver everyting as if we were going to do current and it hads it back as it wasnts.
void newton_aux_ce(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double C=in->C;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;
/*
double S=((in->Rshort+in->Rcontact)-in->Rshunt)/(in->Rshunt*(in->Rshort+in->Rcontact));

*i=i0+C*(V-in->Vapplied_last)/in->dt-(V)*S;
*didv=didv0+C*(1.0)/in->dt-S;
*didphi=didphi0;
*didxil=didxil0;
*didxipl=didxipl0;
*didphir=didphir0;
*didxir=didxir0;
*didxipr=didxipr0;
*/
double S=1.0/(in->Rshort+in->Rcontact);
double L=in->ce_L;
*i=i0+C*(V-in->Vapplied_last)/in->dt+V/in->Rshunt+(V-L*(i0+C*(V-in->Vapplied_last)/in->dt+V/in->Rshunt-in->ilast)/in->dt)*S;
*didv=didv0+C*(1.0)/in->dt+1.0/in->Rshunt+(1.0-L*(didv0+C*(1.0)/in->dt+1.0/in->Rshunt)/in->dt)*S;//didv0+C*(1.0)/in->dt-S;
*didphi=didphi0+(-L*(didphi0)/in->dt)*S;//didphi0;
*didxil=didxil0+(-L*(didxil0)/in->dt)*S;//didxil0;
*didxipl=didxipl0+(-L*(didxipl0)/in->dt)*S;//didxipl0;
*didphir=didphir0+(-L*(didphir0)/in->dt)*S;//didphir0;
*didxir=didxir0+(-L*(didxir0)/in->dt)*S;//didxir0;
*didxipr=didxipr0+(-L*(didxipr0)/in->dt)*S;//didxipr0;


//</clean>
return;
}

double newton_charge_extraction(struct device *in)
{
//<clean>
in->newton_cur=TRUE;
solver_realloc(in);
in->newton_aux=&newton_aux_ce;
solve_all(in);
//</clean>
return 0.0;
}

double newton_sim_tpv_wire(struct device *in)
{
//<clean>
double step=0.001;
double e0;
double e1;
double i0;
double i1;
double deriv;
solve_all(in);
i0=get_I(in);
//getchar();
e0=fabs(i0+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
in->Vapplied+=step;
int count=0;
int max=100;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
//if (e1<clamp) clamp=e1/10.0;
//step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
printf("Tpv error = %e %e\n",in->Vapplied,e1);
if (count>max) break;
count++;
}while(e1>1e-8);
double cur=wire_sim(in->Vapplied);
printf("result = %e %e %e %e\n",i1,in->Vapplied/in->Rshunt,cur,in->Vapplied/cur);
//getchar();
//</clean>
return 0.0;
}

void newton_aux_i(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-i0-V/in->Rshunt;
*didv=-didv0-1.0/in->Rshunt;
*didphi=-didphi0;
*didxil=-didxil0;
*didxipl=-didxipl0;
*didphir=-didphir0;
*didxir=-didxir0;
*didxipr=-didxipr0;
//</clean>
return;
}


double sim_i(struct device *in,double wantedi)
{
//<clean>
in->newton_cur=TRUE;
solver_realloc(in);
glob_wanted=wantedi;
in->newton_aux=&newton_aux_i;
solve_all(in);
//</clean>
return 0.0;
}

void newton_aux_voc(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=V/in->Rshunt-i0;
*didv=1.0/in->Rshunt-didv0;
*didphi=-didphi0;
*didxil=-didxil0;
*didxipl=-didxipl0;
*didphir=-didphir0;
*didxir=-didxir0;
*didxipr=-didxipr0;
//</clean>
return;
}

double sim_voc(struct device *in)
{
//<clean>
in->newton_cur=TRUE;
solver_realloc(in);
in->newton_aux=&newton_aux_voc;
solve_all(in);
printf("Voc=%le\n",get_equiv_V(in));
//</clean>
return 0.0;
}

void newton_aux_externalv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-(V+in->Rcontact*(i0+V/in->Rshunt));
*didv=-(1.0+in->Rcontact*(didv0+1.0/in->Rshunt));
*didphi=-(in->Rcontact*(didphi0));
*didxil=-(in->Rcontact*(didxil0));
*didxipl=-(in->Rcontact*(didxipl0));
*didphir=-(in->Rcontact*(didphir0));
*didxir=-(in->Rcontact*(didxir0));
*didxipr=-(in->Rcontact*(didxipr0));
return;
}

double sim_externalv(struct device *in,double wantedv)
{
//printf("Sim External V=%le Psun=%le\n",wantedv,in->Psun);
in->newton_cur=TRUE;
solver_realloc(in);

glob_wanted=wantedv;
in->newton_aux=&newton_aux_externalv;
solve_all(in);
return 0.0;
}

void newton_aux_tpv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double C=in->C;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;
double Rdrain=1.0/(1.0/in->Rshunt+1.0/(in->Rscope+in->Rcontact));

*i=(i0+C*(V-in->Vapplied_last)/in->dt)-V/Rdrain;
*didv=(didv0+C*(1.0)/in->dt)-1.0/Rdrain;
*didphi=didphi0;
*didxil=didxil0;
*didxipl=didxipl0;
*didphir=didphir0;
*didxir=didxir0;
*didxipr=didxipr0;
//</clean>
return;
}

double newton_sim_tpv_fast(struct device *in)
{
//<clean>
in->newton_cur=TRUE;
solver_realloc(in);

in->newton_aux=&newton_aux_tpv;
solve_all(in);
return get_I(in)+in->C*(in->Vapplied-in->Vapplied_last)+in->Vapplied/in->Rshunt;
//</clean>
}

void solve_all(struct device *in)
{
solve_cur(in);
}

void newton_sim_jv(struct device *in)
{
in->newton_cur=FALSE;
solver_realloc(in);

solve_all(in);
}

void newton_aux_tpc(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
//<clean>
double C=0.0;
double L=0.0;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;
double itot=i0+C*(V-in->Vapplied_last)/in->dt+V/in->Rshunt;
double ditotdv=didv0+C*(1.0)/in->dt+1.0/in->Rshunt;
C=in->C;
L=in->tpc_L;
*i=in->tpc_power_suply+L*(itot-in->ilast)/in->dt-(V+(in->Rcontact+in->Rtpc_short)*(itot));
*didv=L*(ditotdv)/in->dt-(1.0+(in->Rcontact+in->Rtpc_short)*(ditotdv));
*didphi=L*(didphi0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didphi0));
*didxil=L*(didxil0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didxil0));
*didxipl=L*(didxipl0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didxipl0));
*didphir=L*(didphir0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didphir0));
*didxir=L*(didxir0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didxir0));
*didxipr=L*(didxipr0)/in->dt-((in->Rcontact+in->Rtpc_short)*(didxipr0));
//</clean>
return;
}

double newton_sim_tpc_all(struct device *in)
{
in->newton_cur=TRUE;
solver_realloc(in);
in->newton_aux=&newton_aux_tpc;
solve_all(in);//in->C*(in->Vapplied-in->Vapplied_last)/in->dt+
return get_I(in)+in->C*(in->Vapplied-in->Vapplied_last)+in->Vapplied/in->Rshunt;
}

void newton_sim_tpc(struct device *in)
{
in->newton_cur=FALSE;
solver_realloc(in);
solve_all(in);
}

double newton_sim_tpv(struct device *in)
{
//<clean>
printf("Looking for Voc\n");
double C=in->C;
double clamp=0.01;
double step=0.01;
double e0;
double e1;
double i0;
double i1;
double deriv;
double Rdrain=1.0/(1.0/in->Rshunt+1.0/1e6);

solve_all(in);
i0=get_I(in);
e0=fabs(i0-in->Vapplied/Rdrain);

in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1-in->Vapplied/Rdrain);

deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;

int count=0;
int max=100;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1-in->Vapplied/Rdrain);
deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
printf("%d TPV find Voc Vapplied=%le step=%le\n",count,in->Vapplied,step);
if (count>max) break;
count++;

}while(e1>1e-12);

double ret=in->Vapplied-C*(i1-in->Ilast)/in->dt;
return ret;
//</clean>
}

